intro = """
          
██╗░░░░░██╗░░░██╗░█████╗░██╗███████╗███████╗██████╗░░░░░░░███╗░░░███╗██╗░░░░░
██║░░░░░██║░░░██║██╔══██╗██║██╔════╝██╔════╝██╔══██╗░░░░░░████╗░████║██║░░░░░
██║░░░░░██║░░░██║██║░░╚═╝██║█████╗░░█████╗░░██████╔╝█████╗██╔████╔██║██║░░░░░
██║░░░░░██║░░░██║██║░░██╗██║██╔══╝░░██╔══╝░░██╔══██╗╚════╝██║╚██╔╝██║██║░░░░░
███████╗╚██████╔╝╚█████╔╝██║██║░░░░░███████╗██║░░██║░░░░░░██║░╚═╝░██║███████╗
╚══════╝░╚═════╝░░╚════╝░╚═╝╚═╝░░░░░╚══════╝╚═╝░░╚═╝░░░░░░╚═╝░░░░░╚═╝╚══════╝
"""

classifiers = {
    "lr": "Logistic Regression",
    "sgd": "Stochastic Gradient Descent",
    "perc": "Perceptron",
    "pass": "Passive Aggressive Classifier",
    "ridg": "Ridge Classifier",
    "svm": "Support Vector Machine",
    "knn": "K-Nearest Neighbours",
    "dt": "Decision Trees",
    "nb": "Naive Bayes",
    "rfc": "Random Forest Classifier",
    "gbc": "Gradient Boosting Classifier",
    "ada": "AdaBoost Classifier",
    "bag": "Bagging Classifier",
    "extc": "Extra Trees Classifier",
    "lgbm": "LightGBM Classifier",
    "cat": "CatBoost Classifier",
    "xgb": "XGBoost Classifier",
    "ann": "Multi Layer Perceptron Classifier",
}

regressors = {
    "lin": "Linear Regression",
    "sgd": "Stochastic Gradient Descent Regressor",
    "krr": "Kernel Ridge Regressor",
    "elas": "Elastic Net Regressor",
    "br": "Bayesian Ridge Regressor",
    "svr": "Support Vector Regressor",
    "knr": "K-Neighbors Regressor",
    "dt": "Decision Trees Regressor",
    "rfr": "Random Forest Regressor",
    "gbr": "Gradient Boost Regressor",
    "ada": "AdaBoost Regressor",
    "bag": "Bagging Regressor",
    "extr": "Extra Trees Regressor",
    "lgbm": "LightGBM Regressor",
    "xgb": "XGBoost Regressor",
    "cat": "Catboost Regressor",
    "ann": "Multi-Layer Perceptron Regressor",
}

classifiers_ver = {
    "lr": "Logistic Regression",
    "sgd": "Stochastic Gradient Descent",
    "perc": "Perceptron",
    "pass": "Passive Aggressive Classifier",
    "ridg": "Ridge Classifier",
    "svm": "Support Vector Machine",
    "knn": "K-Nearest Neighbours",
    "dt": "Decision Trees",
    "nb": "Naive Bayes",
    "rfc": "Random Forest Classifier",
    "gbc": "Gradient Boosting Classifier",
    "ada": "AdaBoost Classifier",
    "bag": "Bagging Classifier",
    "extc": "Extra Trees Classifier",
    "lgbm": "LightGBM Classifier",
    "cat": "CatBoost Classifier",
    "xgb": "XGBoost Classifier",
    "ann": "Multi Layer Perceptron Classifier",
    "all": "All Classifiers",
}

regressors_ver = {
    "lin": "Linear Regression",
    "sgd": "Stochastic Gradient Descent Regressor",
    "krr": "Kernel Ridge Regressor",
    "elas": "Elastic Net Regressor",
    "br": "Bayesian Ridge Regressor",
    "svr": "Support Vector Regressor",
    "knr": "K-Neighbors Regressor",
    "dt": "Decision Trees Regressor",
    "rfr": "Random Forest Regressor",
    "gbr": "Gradient Boost Regressor",
    "ada": "AdaBoost Regressor",
    "bag": "Bagging Regressor",
    "extr": "Extra Trees Regressor",
    "lgbm": "LightGBM Regressor",
    "xgb": "XGBoost Regressor",
    "cat": "Catboost Regressor",
    "ann": "Multi Layer Perceptron Regressor",
    "all": "All Regressors",
}
params_use_warning = (
    "Params will not work with predictor = 'all'. Settings params = {} "
)

unsupported_pred_warning = """Predictor not available. Please use the predictor which is supported by LuciferML.
Check the documentation for more details.\nConflicting Predictor is : {}"""



import time
from collections import Counter

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from colorama import Fore
from IPython.display import display
from scipy.special import boxcox1p
from scipy.stats import norm, probplot, skew

from luciferml.supervised.utils.configs import intro


class Preprocess:

    def __init__(self, dataset, columns, except_columns=[]):
        self.__dataset = dataset
        self.__columns = columns
        self.__except_columns = except_columns

    def __plotter(self, name, text, color):
        plt.figure(figsize=(20, 10))
        plt.subplot(1, 2, 1)
        sns.distplot(
            self.__dataset[name],
            fit=norm,
            color=color,
            label="Skewness: %.2f" % (self.__dataset[name].skew()),
        )
        plt.title(
            name.capitalize()
            + " Distplot for {} {} Skewness Transformation".format(name, text),
            color="black",
        )
        plt.legend()
        plt.subplot(1, 2, 2)

        probplot(self.__dataset[name], plot=plt)
        plt.show()

    def __skewcheck(self):
        numeric_feats = self.__dataset.dtypes[self.__dataset.dtypes !=
                                              "object"].index
        if not len(self.__except_columns) == 0:
            if len(self.__except_columns) > len(numeric_feats):
                numeric_feats = set(self.__except_columns) - set(numeric_feats)
            else:
                numeric_feats = set(numeric_feats) - set(self.__except_columns)
        skewed_feats = (
            self.__dataset[numeric_feats]
            .apply(lambda x: skew(x.dropna()))
            .sort_values(ascending=False)
        )
        print(Fore.GREEN + "\nSkewness in numerical features: \n")
        skewness = pd.DataFrame(skewed_feats, columns=["Skewness"])
        display(skewness)
        skew_dict = dict(skewness["Skewness"])
        skewed_features = skewness.index
        return (skewed_features, skew_dict)

    def skewcorrect(self) -> pd.DataFrame:
        """
            Plots distplot and probability plot for non-normalized data and after normalizing the provided data.
            Normalizes data using boxcox normalization

        :returns: Scaled Dataset
        :rtype: pd.DataFrame

        Example:

         1) All Columns

                     from luciferml.preprocessing import Preprocess as pp

                     import pandas as pd

                     dataset = pd.read_csv('/examples/Social_Network_Ads.csv')
                     prep = pp(dataset, dataset.columns)
                     dataset = prep.skewcorrect(dataset)

         2) Except column/columns

                     from luciferml.preprocessing import Preprocess as pp

                     import pandas as pd

                     dataset = pd.read_csv('/examples/Social_Network_Ads.csv')
                     prep = pp(dataset, dataset.columns, except_columns=['Purchased'])
                     dataset = prep.skewcorrect()


        """
        try:
            start = time.time()
            print(Fore.MAGENTA + intro, "\n")
            print(Fore.GREEN + "Started LuciferML [", "\u2713", "]\n")
            if not isinstance(self.__dataset, pd.DataFrame):
                print(
                    Fore.RED + "TypeError: This Function expects  Pandas Dataframe but {}".format(
                        type(self.__dataset)
                    ),
                    " is given \n",
                )
                end = time.time()
                print(Fore.GREEN + "Elapsed Time: ", end - start, "seconds\n")
                return

            (skewed_features, skew_dict) = self.__skewcheck()
            for column_name in skewed_features:
                lam = 0
                (mu, sigma) = norm.fit(self.__dataset[column_name])
                print(
                    Fore.CYAN +
                    "Skewness Before Transformation for {}: ".format(
                        column_name),
                    self.__dataset[column_name].skew(),
                    "\n",
                )
                print(
                    Fore.CYAN + "Mean before Transformation for {} : {}, Standard Deviation before Transformation for {} : {}".format(
                        column_name.capitalize(), mu, column_name.capitalize(), sigma
                    ),
                    "\n",
                )
                self.__plotter(
                    column_name, "Before", "lightcoral")
                try:
                    if skew_dict[column_name] > 0.75:
                        lam = 0.15
                    self.__dataset[column_name] = boxcox1p(
                        self.__dataset[column_name], lam)
                    print(
                        Fore.GREEN +
                        "Skewness After Transformation for {}: ".format(
                            column_name),
                        self.__dataset[column_name].skew(),
                        "\n",
                    )
                    (mu, sigma) = norm.fit(self.__dataset[column_name])
                    print(
                        Fore.GREEN + "Mean before Transformation for {} : {}, Standard Deviation before Transformation for {} : {}".format(
                            column_name.capitalize(),
                            mu,
                            column_name.capitalize(),
                            sigma,
                        ),
                        "\n",
                    )
                    self.__plotter(
                        column_name, "After", "orange")
                except Exception as error:
                    print(
                        Fore.RED + "\nPlease check your dataset's column :",
                        column_name,
                        "Raised Error: ",
                        error,
                        "\n",
                    )
                    pass
            end = time.time()
            print(Fore.GREEN + "Elapsed Time: ", end - start, "seconds\n")
            return self.__dataset

        except Exception as error:
            print(Fore.RED + "Skewness Correction Failed with error : ", error, "\n")

    def detect_outliers(self):
        """
        This function takes dataset and columns as input and finds Q1, Q3 and IQR for that list of column
        Detects the outlier and it index and stores them in a list.
        Then it creates as counter object with that list and stores it
        in Multiple Outliers list if the value of outlier is greater than 1.5
        
        Ex:
            1) For printing no. of outliers.
                print("number of outliers detected --> ",
                len(dataset.loc[detect_outliers(dataset, dataset.columns[:-1])]))
            2) Printing rows and columns collecting the outliers
                dataset.loc[detect_outliers(dataset.columns[:-1])]
            3) Dropping those detected outliers
                dataset = dataset.drop(detect_outliers(dataset.columns[:-1]),axis = 0).reset_index(drop = True)
        """
        outlier_indices = []
        for column in self.__columns:
            Q1 = np.percentile(self.__dataset[column], 25)
            Q3 = np.percentile(self.__dataset[column], 75)
            IQR = Q3 - Q1
            outlier_step = IQR * 1.5
            outlier_list_col = self.__dataset[(self.__dataset[column] < Q1 - outlier_step)
                                              | (self.__dataset[column] > Q3 + outlier_step)].index
            outlier_indices.extend(outlier_list_col)
        outlier_indices = Counter(outlier_indices)
        multiple_outliers = list(
            i for i, v in outlier_indices.items() if v > 1.5)
        return multiple_outliers

    def preprocess(self):

        display(self.__datasetdescribe().T.style.bar(
            subset=['mean'],
            color='#606ff2').background_gradient(
            subset=['std'], cmap='PuBu').background_gradient(subset=['50%'], cmap='PuBu'))



import time

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import shap
from imblearn.over_sampling import SMOTE
from luciferml.supervised.utils.configs import *
from sklearn.compose import ColumnTransformer
from sklearn.decomposition import PCA, KernelPCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.inspection import permutation_importance
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, OneHotEncoder, StandardScaler
from colorama import Fore


class PreProcesser:
    def data_preprocess(
        self, features, labels, test_size, random_state, smote, k_neighbors
    ):
        try:
            if smote == "y":
                sm = SMOTE(k_neighbors=k_neighbors, random_state=random_state)
                features, labels = sm.fit_resample(features, labels)
            # Splitting ---------------------------------------------------------------------
            X_train, X_val, y_train, y_val = train_test_split(
                features, labels, test_size=test_size, random_state=random_state
            )
            # Scaling ---------------------------------------------------------------------
            sc = StandardScaler()
            X_train = sc.fit_transform(X_train)
            X_val = sc.transform(X_val)
            return (X_train, X_val, y_train, y_val, sc)
        except Exception as error:
            print(Fore.RED + "Preprocessing Failed with error: ", error, "\n")

    def confusion_matrix(self, y_pred, y_val):
        """
        Takes Predicted data and Validation data as input and prepares and plots Confusion Matrix.
        """
        try:
            cm = confusion_matrix(y_val, y_pred)
            ax = plt.subplot()
            sns.heatmap(cm, annot=True, fmt="g", ax=ax)
            ax.set_xlabel("Predicted labels")
            ax.set_ylabel("True labels")
            ax.set_title("Confusion Matrix")
            ax.xaxis.set_ticklabels(np.unique(y_val))
            ax.yaxis.set_ticklabels(np.unique(y_val))
            plt.show()
        except Exception as error:
            print(
                Fore.RED + "Building Confusion Matrix Failed with error :", error, "\n"
            )

    def dimensionality_reduction(
        self,
        lda,
        pca,
        X_train,
        X_val,
        y_train,
        n_components_lda,
        n_components_pca,
        pca_kernel,
        start,
    ):
        """
        Performs Dimensionality Reduction on Training and Validation independent variables.
        """
        try:
            if lda == "y":
                lda = LDA(n_components=n_components_lda)
                X_train = lda.fit_transform(X_train, y_train)
                X_val = lda.transform(X_val)
            if pca == "y" and not lda == "y":
                if not pca_kernel == "linear":
                    try:

                        kpca = KernelPCA(
                            n_components=n_components_pca, kernel=pca_kernel
                        )
                        X_train = kpca.fit_transform(X_train)
                        X_val = kpca.transform(X_val)
                    except MemoryError as error:
                        print(error)
                        end = time.time()
                        print("Time Elapsed :", end - start)
                        return

                elif pca_kernel == "linear":
                    pca = PCA(n_components=n_components_pca)
                    X_train = pca.fit_transform(X_train)
                    X_val = pca.transform(X_val)
                else:
                    print("Un-identified PCA Kernel\n")
                    return
            return (X_train, X_val)
        except Exception as error:
            print(
                Fore.RED + "Dimensionality Reduction Failed with error :", error, "\n"
            )
            return (X_train, X_val)

    def encoder(self, features, labels):
        """
        Takes features and labels as arguments and encodes features using onehot encoding and labels with label encoding.
        Returns Encoded Features and Labels.
        """
        try:
            cat_features = [
                i for i in features.columns if features.dtypes[i] == "object"
            ]
            if len(cat_features) >= 1:
                index = []
                for i in range(0, len(cat_features)):
                    index.append(features.columns.get_loc(cat_features[i]))
                ct = ColumnTransformer(
                    transformers=[("encoder", OneHotEncoder(), index)],
                    remainder="passthrough",
                )
                print("Encoding Features [*]\n")
                features = np.array(ct.fit_transform(features))
            if labels.dtype == "O":
                le = LabelEncoder()
                labels = le.fit_transform(labels)
            return (features, labels)
        except Exception as error:
            print(Fore.RED + "Encoding Failed with error :", error)

    def permutational_feature_imp(self, features, X_test, y_test, model):
        perm_importance = permutation_importance(model, X_test, y_test)
        sorted_idx = perm_importance.importances_mean.argsort()
        plt.barh(
            features.columns[sorted_idx], perm_importance.importances_mean[sorted_idx]
        )
        plt.xlabel("Feature Importance")

    def shap_feature_imp(self, features, X_train, model, *args, **kwargs):
        explainer = shap.TreeExplainer(model)
        shap_values = explainer.shap_values(X_train)
        shap.summary_plot(
            shap_values,
            X_train,
            feature_names=features.columns,
            plot_type="bar",
            *args,
            **kwargs
        )
        shap.summary_plot(
            shap_values, X_train, feature_names=features.columns, *args, **kwargs
        )
        for i in range(len(features.columns)):
            shap.dependence_plot(
                i, shap_values, X_train, feature_names=features.columns, *args, **kwargs
            )



import traceback

from luciferml.supervised.utils.predictors import classification_predictor
from luciferml.supervised.utils.predictors import regression_predictor
from luciferml.supervised.utils.tuner.optuna.optuna_base import Tuner
from colorama import Fore


def luciferml_tuner(
    predictor,
    objective,
    n_trials,
    sampler,
    direction,
    X_train,
    y_train,
    cv_folds,
    random_state,
    metric,
    all_mode=False,
    verbose=False,
    isReg=False,
):
    """
    Takes classifier, tune-parameters, Training Data and no. of folds as input and Performs GridSearch Crossvalidation.
    """
    tuner = Tuner(n_trials=n_trials, sampler=sampler, direction=direction)
    try:
        best_params, best_score = tuner.tune(objective)
        if isReg:
            model, _ = regression_predictor(
                predictor,
                best_params,
                X_train,
                y_train,
                cv_folds,
                random_state,
                metric,
                mode="tune",
                verbose=verbose,
            )
        if not isReg:
            model, _ = classification_predictor(
                predictor,
                best_params,
                X_train,
                y_train,
                cv_folds,
                random_state,
                metric,
                mode="tune",
                verbose=verbose,
            )
        if not all_mode:
            print(Fore.CYAN + "        Best Params: ", best_params)
            print(Fore.CYAN + "        Best Score: ", best_score * 100, "\n")
        return best_params, best_score, model
    except Exception as error:
        print(Fore.RED + "HyperParam Tuning Failed with Error: ", error, "\n")
        return None, 0, None



from tkinter import N
from catboost import CatBoostClassifier, CatBoostRegressor
from colorama import Fore
from lightgbm import LGBMClassifier, LGBMRegressor
from luciferml.supervised.utils.tuner.optuna.objectives.classification_objectives import (
    ClassificationObjectives,
)
from luciferml.supervised.utils.tuner.optuna.objectives.regression_objectives import (
    RegressionObjectives,
)
from sklearn.ensemble import (
    AdaBoostClassifier,
    AdaBoostRegressor,
    BaggingClassifier,
    BaggingRegressor,
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import (
    BayesianRidge,
    ElasticNet,
    LinearRegression,
    LogisticRegression,
    PassiveAggressiveClassifier,
    Perceptron,
    RidgeClassifier,
    SGDClassifier,
    SGDRegressor,
)
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import SVC, SVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from xgboost import XGBClassifier, XGBRegressor


def classification_predictor(
    predictor,
    params,
    X_train,
    y_train,
    cv_folds,
    random_state,
    metric,
    mode="single",
    verbose=False,
    lgbm_objective="binary",
):
    """
    Takes Predictor string , parameters , Training and Validation set and Returns a classifier for the Choosen Predictor.
    """
    try:
        objective = ClassificationObjectives(
            X_train,
            y_train,
            cv=cv_folds,
            random_state=random_state,
            metric=metric,
            lgbm_objective=lgbm_objective,
        )
        if predictor == "lr":
            if mode == "single":
                print(
                    Fore.YELLOW + "Training Logistic Regression on Training Set [*]\n"
                )
            classifier = LogisticRegression(**params)
            objective_to_be_tuned = objective.lr_classifier_objective

        elif predictor == "sgd":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Stochastic Gradient Descent on Training Set [*]\n"
                )
            classifier = SGDClassifier(**params)
            objective_to_be_tuned = objective.sgd_classifier_objective

        elif predictor == "perc":
            if mode == "single":
                print(Fore.YELLOW + "Training Perceptron on Training Set [*]\n")
            classifier = Perceptron(**params)
            objective_to_be_tuned = objective.perc_classifier_objective

        elif predictor == "pass":
            if mode == "single":
                print(Fore.YELLOW + "Training Passive Aggressive on Training Set [*]\n")
            classifier = PassiveAggressiveClassifier(**params)
            objective_to_be_tuned = objective.pass_classifier_objective

        elif predictor == "ridg":
            if mode == "single":
                print(Fore.YELLOW + "Training Ridge Classifier on Training Set [*]\n")
            classifier = RidgeClassifier(**params)
            objective_to_be_tuned = objective.ridg_classifier_objective

        elif predictor == "svm":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Support Vector Machine on Training Set [*]\n"
                )
            classifier = SVC(**params)
            objective_to_be_tuned = objective.svm_classifier_objective

        elif predictor == "knn":
            if mode == "single":
                print(
                    Fore.YELLOW + "Training K-Nearest Neighbours on Training Set [*]\n"
                )
            classifier = KNeighborsClassifier(**params)
            objective_to_be_tuned = objective.knn_classifier_objective

        elif predictor == "dt":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Decision Tree Classifier on Training Set [*]\n"
                )
            classifier = DecisionTreeClassifier(**params)
            objective_to_be_tuned = objective.dt_classifier_objective

        elif predictor == "nb":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Naive Bayes Classifier on Training Set [*]\n"
                )
            classifier = GaussianNB(**params)
            objective_to_be_tuned = None

        elif predictor == "rfc":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Random Forest Classifier on Training Set [*]\n"
                )
            classifier = RandomForestClassifier(**params)
            objective_to_be_tuned = objective.rfc_classifier_objective

        elif predictor == "gbc":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Gradient Boosting Classifier on Training Set [*]\n"
                )
            classifier = GradientBoostingClassifier(**params)
            objective_to_be_tuned = objective.gbc_classifier_objective

        elif predictor == "ada":
            if mode == "single":
                print(
                    Fore.YELLOW + "Training AdaBoost Classifier on Training Set [*]\n"
                )
            classifier = AdaBoostClassifier(**params)
            objective_to_be_tuned = objective.ada_classifier_objective

        elif predictor == "bag":
            if mode == "single":
                print(Fore.YELLOW + "Training Bagging Classifier on Training Set [*]\n")
            classifier = BaggingClassifier(**params)
            objective_to_be_tuned = objective.bag_classifier_objective

        elif predictor == "extc":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Extra Trees Classifier on Training Set [*]\n"
                )
            classifier = ExtraTreesClassifier(**params)
            objective_to_be_tuned = objective.extc_classifier_objective

        elif predictor == "lgbm":
            if mode == "single":
                print(Fore.YELLOW + "Training LightGBM on Training Set [*]\n")
            classifier = LGBMClassifier(**params)
            objective_to_be_tuned = objective.lgbm_classifier_objective

        elif predictor == "cat":
            if mode == "single":
                print(Fore.YELLOW + "Training CatBoostClassifier on Training Set [*]\n")
            params["verbose"] = verbose
            classifier = CatBoostClassifier(**params)
            params.pop("verbose")
            objective_to_be_tuned = objective.cat_classifier_objective

        elif predictor == "xgb":
            if mode == "single":
                print(Fore.YELLOW + "Training XGBClassifier on Training Set [*]\n")
            if verbose:
                params["verbosity"] = 2
            if not verbose:
                params["verbosity"] = 0
                
            classifier = XGBClassifier(**params)
            params.pop("verbosity")
            objective_to_be_tuned = objective.xgb_classifier_objective

        elif predictor == "ann":
            classifier = MLPClassifier(**params)
            objective_to_be_tuned = objective.mlp_classifier_objective
        return (classifier, objective_to_be_tuned)
    except Exception as error:
        print(Fore.RED + "Model Build Failed with error :", error, "\n")


def regression_predictor(
    predictor,
    params,
    X_train,
    y_train,
    cv_folds,
    random_state,
    metric,
    mode="single",
    verbose=False,
):
    """
    Takes Predictor string , parameters , Training and Validation set and Returns a regressor for the Choosen Predictor.
    """
    try:
        objective = RegressionObjectives(
            X_train, y_train, cv=cv_folds, random_state=random_state, metric=metric
        )
        if predictor == "lin":
            if mode == "single":
                print(Fore.YELLOW + "Training Linear Regression on Training Set [*]\n")
            regressor = LinearRegression(**params)
            objective_to_be_tuned = objective.lin_regressor_objective
        elif predictor == "sgd":
            if mode == "single":
                print(
                    "Training Stochastic Gradient Descent Regressor on Training Set [*]\n"
                )
            regressor = SGDRegressor(**params)
            objective_to_be_tuned = objective.sgd_regressor_objective
        elif predictor == "krr":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Kernel Ridge Regressor on Training Set [*]\n"
                )
            regressor = KernelRidge(**params)
            objective_to_be_tuned = objective.krr_regressor_objective
        elif predictor == "elas":
            if mode == "single":
                print(
                    Fore.YELLOW + "Training ElasticNet Regressor on Training Set [*]\n"
                )
            regressor = ElasticNet(**params)
            objective_to_be_tuned = objective.elas_regressor_objective
        elif predictor == "br":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training BayesianRidge Regressor on Training Set [*]\n"
                )
            regressor = BayesianRidge(**params)
            objective_to_be_tuned = objective.br_regressor_objective
        elif predictor == "svr":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Support Vector Machine on Training Set [*]\n"
                )
            regressor = SVR(**params)
            objective_to_be_tuned = objective.svr_regressor_objective
        elif predictor == "knr":
            if mode == "single":
                print(
                    Fore.YELLOW + "Training KNeighbors Regressor on Training Set [*]\n"
                )
            regressor = KNeighborsRegressor(**params)
            objective_to_be_tuned = objective.knr_regressor_objective
        elif predictor == "dt":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Decision Tree regressor on Training Set [*]\n"
                )
            regressor = DecisionTreeRegressor(**params)
            objective_to_be_tuned = objective.dt_regressor_objective
        elif predictor == "rfr":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Random Forest regressor on Training Set [*]\n"
                )
            regressor = RandomForestRegressor(**params)
            objective_to_be_tuned = objective.rfr_regressor_objective
        elif predictor == "gbr":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Gradient Boosting Regressor  on Training Set [*]\n"
                )
            regressor = GradientBoostingRegressor(**params)
            objective_to_be_tuned = objective.gbr_regressor_objective

        elif predictor == "ada":
            if mode == "single":
                print(Fore.YELLOW + "Training AdaBoost Regressor on Training Set [*]\n")
            regressor = AdaBoostRegressor(**params)
            objective_to_be_tuned = objective.ada_regressor_objective
        elif predictor == "bag":
            if mode == "single":
                print(Fore.YELLOW + "Training Bagging Regressor on Training Set [*]\n")
            regressor = BaggingRegressor(**params)
            objective_to_be_tuned = objective.bag_regressor_objective
        elif predictor == "extr":
            if mode == "single":
                print(
                    Fore.YELLOW + "Training Extra Trees Regressor on Training Set [*]\n"
                )
            regressor = ExtraTreesRegressor(**params)
            objective_to_be_tuned = objective.extr_regressor_objective
        elif predictor == "xgb":
            if mode == "single":
                print(Fore.YELLOW + "Training XGBregressor on Training Set [*]\n")
            regressor = XGBRegressor(**params)
            objective_to_be_tuned = objective.xgb_regressor_objective
        elif predictor == "lgbm":
            if mode == "single":
                print(Fore.YELLOW + "Training LGBMRegressor on Training Set [*]\n")
            regressor = LGBMRegressor(**params)
            objective_to_be_tuned = objective.lgbm_regressor_objective
        elif predictor == "cat":
            if mode == "single":
                print(Fore.YELLOW + "Training CatBoost Regressor on Training Set [*]\n")
            params["verbose"] = verbose
            regressor = CatBoostRegressor(**params)
            params.pop("verbose")
            objective_to_be_tuned = objective.cat_regressor_objective
        elif predictor == "ann":
            if mode == "single":
                print(
                    Fore.YELLOW
                    + "Training Multi Layered Perceptron on Training Set [*]\n"
                )
            regressor = MLPRegressor(**params)
            objective_to_be_tuned = objective.mlp_regressor_objective
        return (regressor, objective_to_be_tuned)
    except Exception as error:
        print(Fore.RED + "Model Build Failed with error :", error, "\n")



from catboost import CatBoostClassifier
from lightgbm import LGBMClassifier
from sklearn.ensemble import (
    AdaBoostClassifier,
    BaggingClassifier,
    ExtraTreesClassifier,
    GradientBoostingClassifier,
    RandomForestClassifier,
)
from sklearn.linear_model import (
    LogisticRegression,
    PassiveAggressiveClassifier,
    Perceptron,
    RidgeClassifier,
    SGDClassifier,
)
from sklearn.model_selection import cross_val_score
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier


class ClassificationObjectives:
    def __init__(
        self, X, y, cv=5, random_state=42, metric="accuracy", lgbm_objective="binary"
    ):
        self.metric = metric
        self.cv = cv
        self.X = X
        self.y = y
        self.random_state = random_state
        self.lgbm_objective = lgbm_objective

    def lr_classifier_objective(self, trial):
        param = {
            "C": trial.suggest_loguniform("C", 1e-5, 1e5),
            "solver": trial.suggest_categorical(
                "solver", ["newton-cg", "lbfgs", "liblinear", "sag", "saga"]
            ),
            "max_iter": trial.suggest_int("max_iter", 1, 1000),
            "tol": trial.suggest_loguniform("tol", 1e-5, 1e-2),
            "random_state": self.random_state,
        }
        clf = LogisticRegression(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def sgd_classifier_objective(self, trial):
        param = {
            "loss": trial.suggest_categorical(
                "loss",
                ["hinge", "log", "modified_huber", "squared_hinge", "perceptron"],
            ),
            "penalty": trial.suggest_categorical("penalty", ["l2", "l1", "elasticnet"]),
            "alpha": trial.suggest_loguniform("alpha", 1e-5, 1e5),
            "l1_ratio": trial.suggest_uniform("l1_ratio", 0, 1),
            "fit_intercept": trial.suggest_categorical("fit_intercept", [True, False]),
            "max_iter": trial.suggest_int("max_iter", 1, 1000),
            "tol": trial.suggest_loguniform("tol", 1e-5, 1e-2),
            "random_state": self.random_state,
        }
        clf = SGDClassifier(**param)
        scores = cross_val_score(
            clf, self.X, self.y, cv=self.cv, scoring=self.metric, n_jobs=-1
        )
        return scores.mean()

    def ridg_classifier_objective(self, trial):
        param = {
            "alpha": trial.suggest_loguniform("alpha", 1e-5, 1e5),
            "fit_intercept": trial.suggest_categorical("fit_intercept", [True, False]),
            "max_iter": trial.suggest_int("max_iter", 1, 1000),
            "tol": trial.suggest_loguniform("tol", 1e-5, 1e-2),
            "random_state": self.random_state,
        }
        clf = RidgeClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def perc_classifier_objective(self, trial):
        param = {
            "penalty": trial.suggest_categorical("penalty", ["l2", "l1", "elasticnet"]),
            "alpha": trial.suggest_loguniform("alpha", 1e-5, 1e5),
            "fit_intercept": trial.suggest_categorical("fit_intercept", [True, False]),
            "max_iter": trial.suggest_int("max_iter", 1, 1000),
            "tol": trial.suggest_loguniform("tol", 1e-5, 1e-2),
            "random_state": self.random_state,
        }
        clf = Perceptron(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def pass_classifier_objective(self, trial):
        param = {
            "C": trial.suggest_loguniform("C", 1e-5, 1e5),
            "fit_intercept": trial.suggest_categorical("fit_intercept", [True, False]),
            "max_iter": trial.suggest_int("max_iter", 1, 1000),
            "tol": trial.suggest_loguniform("tol", 1e-5, 1e-2),
            "random_state": self.random_state,
        }
        clf = PassiveAggressiveClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def svm_classifier_objective(self, trial):
        param = {
            "C": trial.suggest_loguniform("C", 1e-5, 1e5),
            "kernel": trial.suggest_categorical(
                "kernel", ["rbf", "linear", "poly", "sigmoid"]
            ),
            "gamma": trial.suggest_loguniform("gamma", 1e-5, 1e5),
            "degree": trial.suggest_int("degree", 1, 10),
            "coef0": trial.suggest_loguniform("coef0", 1e-5, 1e5),
            "shrinking": trial.suggest_categorical("shrinking", [True, False]),
            "tol": trial.suggest_loguniform("tol", 1e-5, 1e-2),
            "random_state": self.random_state,
        }
        clf = SVC(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def knn_classifier_objective(self, trial):
        param = {
            "n_neighbors": trial.suggest_int("n_neighbors", 1, 256),
            "weights": trial.suggest_categorical("weights", ["uniform", "distance"]),
            "p": trial.suggest_int("p", 1, 10),
            "n_jobs": -1,
            "rows_limit": 100000,
        }
        clf = KNeighborsClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def dt_classifier_objective(self, trial):
        param = {
            "criterion": trial.suggest_categorical("criterion", ["gini", "entropy"]),
            "splitter": trial.suggest_categorical("splitter", ["best", "random"]),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "min_samples_split": trial.suggest_int("min_samples_split", 1, 10),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 10),
            "min_weight_fraction_leaf": trial.suggest_float(
                "min_weight_fraction_leaf", 0, 0.5
            ),
            "max_features": trial.suggest_categorical(
                "max_features", ["auto", "sqrt", "log2"]
            ),
            "random_state": self.random_state,
        }
        clf = DecisionTreeClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def rfc_classifier_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 1, 100),
            "criterion": trial.suggest_categorical("criterion", ["gini", "entropy"]),
            "min_weight_fraction_leaf": trial.suggest_float(
                "min_weight_fraction_leaf", 0, 0.5
            ),
            "max_depth": trial.suggest_int("max_depth", 2, 32),
            "min_samples_split": trial.suggest_int("min_samples_split", 2, 100),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 100),
            "max_features": trial.suggest_float("max_features", 0.01, 1),
            "seed":  self.random_state,
            "n_jobs": -1,
            "max_steps": 10,
        }
        clf = RandomForestClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def gbc_classifier_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 1, 100),
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-5, 1e5),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "min_samples_split": trial.suggest_int("min_samples_split", 1, 10),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 10),
            "min_weight_fraction_leaf": trial.suggest_float(
                "min_weight_fraction_leaf", 0, 0.5
            ),
            "max_features": trial.suggest_categorical(
                "max_features", ["auto", "sqrt", "log2"]
            ),
        }
        clf = GradientBoostingClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def ada_classifier_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 1, 100),
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-5, 1e5),
            "algorithm": trial.suggest_categorical("algorithm", ["SAMME", "SAMME.R"]),
            "random_state": self.random_state,
        }
        clf = AdaBoostClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def bag_classifier_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 1, 100),
            "bootstrap": trial.suggest_categorical("bootstrap", [True, False]),
            "bootstrap_features": trial.suggest_categorical(
                "bootstrap_features", [True, False]
            ),
            "max_samples": trial.suggest_uniform("max_samples", 0.1, 1),
            "max_features": trial.suggest_uniform("max_features", 0.1, 1),
            "n_jobs": -1,
            "random_state": self.random_state,
        }
        clf = BaggingClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def extc_classifier_objective(self, trial):
        param = {
            "criterion": trial.suggest_categorical("criterion", ["gini", "entropy"]),
            "min_weight_fraction_leaf": trial.suggest_float(
                "min_weight_fraction_leaf", 0, 0.5
            ),
            "max_depth": trial.suggest_int("max_depth", 2, 32),
            "min_samples_split": trial.suggest_int("min_samples_split", 2, 100),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 100),
            "max_features": trial.suggest_float("max_features", 0.01, 1),
            "random_state": self.random_state,
            "n_jobs": -1,
            "max_steps": 10,
        }
        clf = ExtraTreesClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def lgbm_classifier_objective(self, trial):
        param = {
            "learning_rate": trial.suggest_categorical(
                "learning_rate", [0.0125, 0.025, 0.05, 0.1]
            ),
            "num_leaves": trial.suggest_int("num_leaves", 2, 2048),
            "lambda_l1": trial.suggest_float("lambda_l1", 1e-8, 10.0, log=True),
            "lambda_l2": trial.suggest_float("lambda_l2", 1e-8, 10.0, log=True),
            "feature_fraction": min(
                trial.suggest_float("feature_fraction", 0.3, 1.0 + 1e-8), 1.0
            ),
            "bagging_fraction": min(
                trial.suggest_float("bagging_fraction", 0.3, 1.0 + 1e-8), 1.0
            ),
            "bagging_freq": trial.suggest_int("bagging_freq", 1, 7),
            "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 1, 100),
            "extra_trees": trial.suggest_categorical("extra_trees", [True, False]),
            "feature_pre_filter": False,
            "boosting_type": "gbdt",
            "seed": self.random_state,
            "num_threads": -1,
            "objective": self.lgbm_objective,
        }
        clf = LGBMClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def cat_classifier_objective(self, trial):
        param = {
            "objective": trial.suggest_categorical("objective", ["Logloss", "CrossEntropy"]),
            "iterations": 1000,
            "colsample_bylevel": trial.suggest_float("colsample_bylevel", 0.01, 0.1),
            "depth": trial.suggest_int("depth", 1, 12),
            "boosting_type": trial.suggest_categorical("boosting_type", ["Ordered", "Plain"]),
            "bootstrap_type": trial.suggest_categorical(
                "bootstrap_type", ["Bayesian", "Bernoulli", "MVS"]
            ),
            "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 1, 100),
            "learning_rate": trial.suggest_categorical(
                    "learning_rate", [0.05, 0.1, 0.2]
                ),
            "rsm": trial.suggest_float("rsm", 0.1, 1),
            "l2_leaf_reg": trial.suggest_float(
                "l2_leaf_reg", 0.0001, 10.0, log=False
            ),
            "random_state": self.random_state,
            "verbose": False,
            "allow_writing_files": False,
        }

        if param["bootstrap_type"] == "Bayesian":
            param["bagging_temperature"] = trial.suggest_float("bagging_temperature", 0, 10)
        elif param["bootstrap_type"] == "Bernoulli":
            param["subsample"] = trial.suggest_float("subsample", 0.1, 1)
        clf = CatBoostClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def xgb_classifier_objective(self, trial):
        param = {
            "learning_rate": trial.suggest_categorical(
                "learning_rate", [0.05, 0.1, 0.2]
            ),
            "eta": trial.suggest_categorical("eta", [0.0125, 0.025, 0.05, 0.1]),
            "max_depth": trial.suggest_int("max_depth", 2, 12),
            "lambda": trial.suggest_float("lambda", 1e-8, 10.0, log=True),
            "alpha": trial.suggest_float("alpha", 1e-8, 10.0, log=True),
            "colsample_bytree": min(
                trial.suggest_float("colsample_bytree", 0.3, 1.0 + 1e-8), 1.0
            ),
            "subsample": min(trial.suggest_float("subsample", 0.3, 1.0 + 1e-8), 1.0),
            "min_child_weight": trial.suggest_int("min_child_weight", 1, 100),
            "tree_method": "hist",
            "booster": "gbtree",
            "n_jobs": -1,
            "seed": self.random_state,
            "verbosity": 0,
            
        }
        clf = XGBClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()

    def mlp_classifier_objective(self, trial):
        param = {
            "hidden_layer_sizes": trial.suggest_int("hidden_layer_sizes", 1, 10),
            "max_iter": trial.suggest_int("max_iter", 1, 2000),
            "dense_1_size": trial.suggest_int("dense_1_size", 4, 100),
            "dense_2_size": trial.suggest_int("dense_2_size", 2, 100),
            "learning_rate": trial.suggest_categorical(
                "learning_rate", [0.005, 0.01, 0.05, 0.1, 0.2]
            ),
            "learning_rate_type": trial.suggest_categorical(
                "learning_rate_type", ["constant", "adaptive"]
            ),
            "alpha": trial.suggest_float("alpha", 1e-8, 10.0, log=True),
            "seed": self.random_state,
        }
        clf = MLPClassifier(**param)
        scores = cross_val_score(
            clf,
            self.X,
            self.y,
            cv=self.cv,
            scoring=self.metric,
            n_jobs=-1,
        )
        return scores.mean()



import copy
import os
import time
import warnings
from pickle import dump, load

import numpy as np
import optuna
import pandas as pd
from colorama import Fore
from IPython.display import display
from luciferml.supervised.utils.best import Best
from luciferml.supervised.utils.configs import *
from luciferml.supervised.utils.predictors import regression_predictor
from luciferml.supervised.utils.preprocesser import PreProcesser
from luciferml.supervised.utils.tuner.luciferml_tuner import luciferml_tuner
from luciferml.supervised.utils.validator import *
from optuna.samplers._tpe.sampler import TPESampler
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score


class Regression:
    def __init__(
        self,
        predictor=["lin"],
        params={},
        tune=False,
        test_size=0.2,
        cv_folds=10,
        random_state=42,
        pca_kernel="linear",
        n_components_lda=1,
        lda="n",
        pca="n",
        n_components_pca=2,
        loss="mean_squared_error",
        validation_split=0.20,
        smote="n",
        k_neighbors=1,
        verbose=False,
        exclude_models=[],
        path=None,
        optuna_sampler=TPESampler(multivariate=True),
        optuna_direction="maximize",
        optuna_n_trials=100,
        optuna_metric="r2",
    ):
        """
        Encodes Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2\n
        scales X_train, X_val using StandardScaler.\n
        Fits every model on training set and predicts results,Finds R2 Score and mean square error\n
        finds accuracy of model applies K-Fold Cross Validation\n
        and stores its accuracies in a dictionary containing Model name as Key and accuracies as values and returns it\n
        Applies HyperParam Tuning and gives best params and accuracy.\n

        Parameters:

            features : array
                        features array
            lables : array
                        labels array
            predictor : list
                        Predicting models to be used
                        Default ['lin'] - 'Linear Regression'\n
                        Available Predictors:
                                lin  - Linear Regression\n
                                sgd  - Stochastic Gradient Descent Regressor\n
                                elas - Elastic Net Regressor\n
                                krr  - Kernel Ridge Regressor\n
                                br   - Bayesian Ridge Regressor\n
                                svr  - Support Vector Regressor\n
                                knr  - K-Nearest Regressor\n
                                dt   - Decision Trees\n
                                rfr  - Random Forest Regressor\n
                                gbr  - Gradient Boost Regressor\n
                                ada  - AdaBoost Regressor,\n
                                bag  - Bagging Regressor,\n
                                extr - Extra Trees Regressor,\n
                                lgbm - LightGB Regressor\n
                                xgb  - XGBoost Regressor\n
                                cat  - Catboost Regressor\n
                                ann  - Multi Layer Perceptron Regressor\n
                                all  - Applies all above regressors\n
            params : dict
                        contains parameters for model
            tune : boolean
                    when True Applies GridSearch CrossValidation
                    Default is False
            test_size: float or int, default=.2
                        If float, should be between 0.0 and 1.0 and represent
                        the proportion of the dataset to include in
                        the test split.
                        If int, represents the absolute number of test samples.
            cv_folds : int
                    No. of cross validation folds. Default = 10
            pca : str
                if 'y' will apply PCA on Train and Validation set. Default = 'n'
            lda : str
                if 'y' will apply LDA on Train and Validation set. Default = 'n'
            pca_kernel : str
                    Kernel to be use in PCA. Default = 'linear'
            n_components_lda : int
                    No. of components for LDA. Default = 1
            n_components_pca : int
                    No. of components for PCA. Default = 2
            loss : str
                    loss method for ann. Default = 'mean_squared_error'
            smote : str,
                Whether to apply SMOTE. Default = 'y'
            k_neighbors : int
                No. of neighbours for SMOTE. Default = 1
            verbose : boolean
                Verbosity of models. Default = False
            exclude_models : list
                List of models to be excluded when using predictor = 'all' . Default = []
            path : list
                List containing path to saved model and scaler. Default = None
                Example: [model.pkl, scaler.pkl]
            random_state : int
                Random random_state for reproducibility. Default = 42
            optuna_sampler : Function
                Sampler to be used in optuna. Default = TPESampler()
            optuna_direction : str
                Direction of optimization. Default = 'maximize'
                Available Directions:
                    maximize : Maximize
                    minimize : Minimize
            optuna_n_trials : int
                No. of trials for optuna. Default = 100
            optuna_metric: str
                Metric to be used in optuna. Default = 'r2'
        Returns:

            Dict Containing Name of Regressor, Its K-Fold Cross Validated Accuracy, RMSE, Prediction set

            Dataframe containing all the models and their accuracies when predictor is 'all'

        Example:

            from luciferml.supervised.regression import Regression

            dataset = pd.read_excel('examples\Folds5x2_pp.xlsx')

            X = dataset.iloc[:, :-1]

            y = dataset.iloc[:, -1]

            regressor = Regression(predictor = 'lin')

            regressor.fit(X, y)

            result = regressor.result()

        """
        self.preprocess = PreProcesser()
        if type(predictor) == list:
            if not "all" in predictor:
                self.predictor = predictor[0] if len(predictor) == 1 else predictor
            else:
                self.predictor = predictor
        else:
            self.predictor = predictor
        bool_pred, pred = pred_check(predictor, pred_type="regression")
        if not bool_pred:
            raise ValueError(unsupported_pred_warning.format(pred))
        self.original_predictor = predictor
        self.params = params
        self.tune = tune
        self.test_size = test_size
        self.cv_folds = cv_folds
        self.random_state = random_state
        self.pca_kernel = pca_kernel
        self.n_components_lda = n_components_lda
        self.lda = lda
        self.pca = pca
        self.n_components_pca = n_components_pca
        self.loss = loss
        self.validation_split = validation_split
        self.rerun = False
        self.smote = smote
        self.k_neighbors = k_neighbors
        self.verbose = verbose
        self.exclude_models = exclude_models
        self.sampler = optuna_sampler
        self.direction = optuna_direction
        self.n_trials = optuna_n_trials
        self.metric = optuna_metric

        self.accuracy_scores = {}
        self.reg_result = {}
        self.rm_squared_error = 0
        self.accuracy = 0
        self.y_pred = []
        self.kfold_accuracy = 0
        self.regressor_name = ""
        self.sc = 0

        self.kfoldacc = []
        self.acc = []
        self.mae = []
        self.rmse = []
        self.bestacc = []
        self.bestparams = []
        self.regressor_model = []
        self.tuned_trained_model = []
        self.best_regressor_path = ""
        self.scaler_path = ""
        self.result_df = pd.DataFrame(index=None)
        self.regressors = copy.deepcopy(regressors)
        for i in self.exclude_models:
            self.regressors.pop(i)
        self.best_regressor = "First Run the Predictor in All mode"
        self.objective = None
        self.pred_mode = ""
        self.model_to_predict = None

        if path != None:
            try:
                self.regressor, self.sc = self.__load(path)
            except Exception as e:
                print(Fore.RED + e)
                print(Fore.RED + "Model not found")
        if not self.verbose:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

    def fit(self, features, labels):
        """[Takes Features and Labels and Encodes Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2
        scales X_train, X_val using StandardScaler.
        Fits model on training set and predicts results, Finds R2 Score and mean square error
        finds accuracy of model applies K-Fold Cross Validation
        and stores its accuracies in a dictionary containing Model name as Key and accuracies as values and returns it
        Applies GridSearch Cross Validation and gives best params out from param list.]

        Args:

            features ([Pandas DataFrame]): [DataFrame containing Features]
            labels ([Pandas DataFrame]): [DataFrame containing Labels]
        """

        self.features = features
        self.labels = labels

        # Time Function ---------------------------------------------------------------------

        self.start = time.time()
        print(Fore.MAGENTA + intro, "\n")
        print(Fore.GREEN + "Started LuciferML [", "\u2713", "]\n")
        if not self.rerun:
            # CHECKUP ---------------------------------------------------------------------
            if not isinstance(self.features, pd.DataFrame) and not isinstance(
                self.labels, pd.Series
            ):
                print(
                    Fore.RED
                    + "TypeError: This Function take features as Pandas Dataframe and labels as Pandas Series. Please check your implementation.\n"
                )
                end = time.time()
                print(self.end - self.start)
                return
            print(Fore.YELLOW + "Preprocessing Started [*]\n")

            self.features, self.labels = self.preprocess.encoder(
                self.features, self.labels
            )
            self.features, self.labels = sparse_check(self.features, self.labels)
            (
                self.X_train,
                self.X_val,
                self.y_train,
                self.y_val,
                self.sc,
            ) = self.preprocess.data_preprocess(
                self.features,
                self.labels,
                self.test_size,
                self.random_state,
                self.smote,
                self.k_neighbors,
            )
            self.X_train, self.X_val = self.preprocess.dimensionality_reduction(
                self.lda,
                self.pca,
                self.X_train,
                self.X_val,
                self.y_train,
                self.n_components_lda,
                self.n_components_pca,
                self.pca_kernel,
                self.start,
            )

        print(Fore.GREEN + "Preprocessing Done [", "\u2713", "]\n")

        if self.original_predictor == "all" or type(self.predictor) == list:
            if 'all' in self.predictor and type(self.predictor)==list:
                self.predictor.remove('all')
            self.model_to_predict = (
                self.predictor if len(self.predictor) > 1 and type(self.predictor) == list else self.regressors
            )
            self.result_df["Name"] = (
                list(self.regressors[i] for i in self.predictor)
                if type(self.predictor) == list and len(self.predictor) > 1
                else list(self.regressors.values())
            )
            self.pred_mode = "all" if type(self.predictor) == list and len(
                self.predictor) > 1 else "single"
            self.__fitall()
            return
        self.regressor, self.objective = regression_predictor(
            self.predictor,
            self.params,
            self.X_train,
            self.y_train,
            self.cv_folds,
            self.random_state,
            self.metric,
            verbose=self.verbose,
        )
        try:
            self.regressor.fit(self.X_train, self.y_train)
        except Exception as error:
            print(Fore.RED + "Regressor Build Failed with error: ", error, "\n")
        finally:
            print(Fore.GREEN + "Model Trained Successfully [", "\u2713", "]\n")

        try:
            print(Fore.YELLOW + "Evaluating Model Performance [*]\n")
            self.y_pred = self.regressor.predict(self.X_val)
            self.accuracy = r2_score(self.y_val, self.y_pred)
            self.m_absolute_error = mean_absolute_error(self.y_val, self.y_pred)
            self.rm_squared_error = mean_squared_error(
                self.y_val, self.y_pred, squared=False
            )
            print(
                Fore.CYAN
                + "        Validation R2 Score is {:.2f} %".format(self.accuracy * 100)
            )
            print(
                Fore.CYAN + "        Validation Mean Absolute Error is :",
                self.m_absolute_error,
            )
            print(
                Fore.CYAN + "        Validation Root Mean Squared Error is :",
                self.rm_squared_error,
            )
            self.regressor_name, self.kfold_accuracy = kfold(
                self.regressor,
                self.predictor,
                self.X_train,
                self.y_train,
                self.cv_folds,
                isReg=True,
            )
        except Exception as error:
            print(Fore.RED + "Model Evaluation Failed with error: ", error, "\n")
        finally:
            print(Fore.GREEN + "Model Evaluation Completed [", "\u2713", "]\n")

        if not self.predictor == "nb" and self.tune:
            self.__tuner()

        print(Fore.GREEN + "Completed LuciferML Run [", "\u2713", "]\n")
        self.end = time.time()
        final_time = self.end - self.start
        print(Fore.BLUE + "Time Elapsed : ", f"{final_time:.2f}", "seconds \n")

    def __fitall(self):
        print(Fore.YELLOW + "Training LuciferML [*]\n")
        if self.params != {}:
            warnings.warn(params_use_warning, UserWarning)
            self.params = {}
        for _, self.predictor in enumerate(self.model_to_predict):
            if not self.predictor in self.exclude_models:
                (self.regressor, self.objective,) = regression_predictor(
                    self.predictor,
                    self.params,
                    self.X_train,
                    self.y_train,
                    self.cv_folds,
                    self.random_state,
                    self.metric,
                    mode="multi",
                    verbose=self.verbose,
                )
                try:
                    self.regressor.fit(self.X_train, self.y_train)
                except Exception as error:
                    print(
                        Fore.RED + regressors[self.predictor],
                        "Model Train Failed with error: ",
                        error,
                        "\n",
                    )
                try:
                    self.y_pred = self.regressor.predict(self.X_val)
                    self.accuracy = r2_score(self.y_val, self.y_pred)
                    self.m_absolute_error = mean_absolute_error(self.y_val, self.y_pred)
                    self.rm_squared_error = mean_squared_error(
                        self.y_val, self.y_pred, squared=False
                    )
                    self.acc.append(self.accuracy * 100)
                    self.rmse.append(self.rm_squared_error)
                    self.mae.append(self.m_absolute_error)
                    self.regressor_name, self.kfold_accuracy = kfold(
                        self.regressor,
                        self.predictor,
                        self.X_train,
                        self.y_train,
                        self.cv_folds,
                        all_mode=True,
                        isReg=True,
                    )
                    self.kfoldacc.append(self.kfold_accuracy)
                    self.regressor_model.append(self.regressor)
                except Exception as error:
                    print(
                        Fore.RED + regressors[self.predictor],
                        "Evaluation Failed with error: ",
                        error,
                        "\n",
                    )

                if self.tune:
                    self.__tuner(all_mode=True, single_mode=False)
                if self.predictor == "nb":
                    self.best_params = ""
                    self.best_accuracy = self.kfold_accuracy
        self.result_df["R2 Score"] = self.acc
        self.result_df["Mean Absolute Error"] = self.mae
        self.result_df["Root Mean Squared Error"] = self.rmse
        self.result_df["KFold Accuracy"] = self.kfoldacc
        self.result_df["Model"] = self.regressor_model

        if self.tune:
            self.result_df["Best Parameters"] = self.bestparams
            self.result_df["Best Accuracy"] = self.bestacc
            self.result_df["Trained Model"] = self.tuned_trained_model
            self.best_regressor = Best(
                self.result_df.loc[self.result_df["Best Accuracy"].idxmax()],
                self.tune,
                isReg=True,
            )
        else:
            self.best_regressor = Best(
                self.result_df.loc[self.result_df["KFold Accuracy"].idxmax()],
                self.tune,
                isReg=True,
            )
        print(Fore.GREEN + "Training Done [", "\u2713", "]\n")
        print(Fore.CYAN + "Results Below\n")
        display(self.result_df)
        print(Fore.GREEN + "\nCompleted LuciferML Run [", "\u2713", "]\n")
        if len(self.model_to_predict) > 1:
            self.best_regressor_path, self.scaler_path = self.save(
                best=True, model=self.best_regressor.model, scaler=self.sc
            )
            print(
                Fore.CYAN
                + "Saved Best Model to {} and its scaler to {}".format(
                    self.best_regressor_path, self.scaler_path
                ),
                "\n",
            )
        self.end = time.time()
        final_time = self.end - self.start
        print(Fore.BLUE + "Time Elapsed : ", f"{final_time:.2f}", "seconds \n")
        return

    def __tuner(self, all_mode=False, single_mode=True):
        if not all_mode or single_mode:
            print(Fore.YELLOW + "Tuning Started [*]\n")
        if not self.predictor == "nb":
            (
                self.best_params,
                self.best_accuracy,
                self.best_trained_model,
            ) = luciferml_tuner(
                self.predictor,
                self.objective,
                self.n_trials,
                self.sampler,
                self.direction,
                self.X_train,
                self.y_train,
                self.cv_folds,
                self.random_state,
                self.metric,
                all_mode=all_mode,
                isReg=True,
            )
        if self.predictor == "nb":
            self.best_params = "Not Applicable"
            self.best_accuracy = 0
        self.bestparams.append(self.best_params)
        self.bestacc.append(self.best_accuracy * 100)
        self.tuned_trained_model.append(self.best_trained_model)
        if not all_mode or single_mode:
            print(Fore.GREEN + "Tuning Done [", "\u2713", "]\n")

    def result(self):
        """[Makes a dictionary containing Regressor Name, K-Fold CV Accuracy, RMSE, Prediction set.]

        Returns:

            [dict]: [Dictionary containing :
                        - "Regressor" - Regressor Name
                        - "Accuracy" - KFold CV Accuracy
                        - "RMSE" - Root Mean Square
                        - "YPred" - Array for Prediction set
                        ]
            [dataframe] : [Dataset containing accuracy and best_params
                            for all predictors only when predictor = 'all' is used
                            ]
        """
        if not self.pred_mode == "all":
            self.reg_result["Regressor"] = self.regressor_name
            self.reg_result["Accuracy"] = self.kfold_accuracy
            self.reg_result["RMSE"] = self.rm_squared_error
            self.reg_result["YPred"] = self.y_pred

            return self.reg_result
        if self.pred_mode == "all":
            return self.result_df

    def predict(self, X_test):
        """[Takes test set and returns predictions for that test set]

        Args:
            X_test ([Array]): [Array Containing Test Set]

        Returns:
            [Array]: [Predicted set for given test set]
        """
        if not self.pred_mode == "all":
            X_test = np.array(X_test)
            if X_test.ndim == 1:
                X_test = X_test.reshape(1, -1)

            y_test = self.regressor.predict(self.sc.transform(X_test))

            return y_test
        if self.pred_mode == "all":
            raise TypeError("Predict is only applicable on single predictor")

    def save(self, path=None, best=False, **kwargs):
        """
        Saves the model and its scaler to a file provided with a path.
        If no path is provided will create a directory named
        lucifer_ml_info/models/ and lucifer_ml_info/scaler/ in current working directory

        Args:

            path ([list]): [List containing path to save the model and scaler.]
                Example: path = ["model.pkl", "scaler.pkl"]

        Returns:

            Path to the saved model and its scaler.
        """
        if not type(path) == list and path != None:
            raise TypeError("Path must be a list")
        if self.pred_mode == "all" and best == False:
            raise TypeError("Cannot save model for all predictors")
        dir_path_model = path[0] if path else "lucifer_ml_info/models/regression/"
        dir_path_scaler = path[1] if path else "lucifer_ml_info/scalers/regression/"
        model_name = regressors[self.predictor].replace(" ", "_")
        if best:
            dir_path_model = "lucifer_ml_info/best/regression/models/"
            dir_path_scaler = "lucifer_ml_info/best/regression/scalers/"
            model_name = self.best_regressor.name.replace(" ", "_")
        os.makedirs(dir_path_model, exist_ok=True)
        os.makedirs(dir_path_scaler, exist_ok=True)
        timestamp = str(int(time.time()))
        path_model = dir_path_model + model_name + "_" + timestamp + ".pkl"
        path_scaler = (
            dir_path_scaler + model_name + "_" + "Scaler" + "_" + timestamp + ".pkl"
        )
        if (
            not kwargs.get("model")
            and not kwargs.get("best")
            and not kwargs.get("scaler")
        ):
            dump(self.regressor, open(path_model, "wb"))
            dump(self.sc, open(path_scaler, "wb"))
        else:
            dump(kwargs.get("model"), open(path_model, "wb"))
            dump(kwargs.get("scaler"), open(path_scaler, "wb"))
        if not best:
            print("Model Saved at {} and Scaler at {}".format(path_model, path_scaler))
        return path_model, path_scaler

    def __load(self, path=None):
        """
        Loads model and scaler from the specified path

        Args:

            path ([list]): [List containing path to load the model and scaler.]
                Example: path = ["model.pkl", "scaler.pkl"]

        Returns:
            [Model] : [Loaded model]
            [Scaler] : [Loaded scaler]
        """
        model_path = path[0] if path[0] else None
        scaler_path = path[1] if path[1] else None
        if not ".pkl" in model_path and not model_path == None:
            raise TypeError(
                "[Error] Model Filetype not supported. Please use .pkl type "
            )
        if not ".pkl" in scaler_path and not scaler_path == None:
            raise TypeError(
                "[Error] Scaler Filetype not supported. Please use .pkl type "
            )
        if model_path != None and scaler_path != None:
            model = load(open(model_path, "rb"))
            scaler = load(open(scaler_path, "rb"))
            print(
                Fore.GREEN
                + "[Info] Model and Scaler Loaded from {} and {}".format(
                    model_path, scaler_path
                )
            )
            return model, scaler
        elif model_path != None and scaler_path == None:
            model = load(open(model_path, "rb"))
            print(Fore.GREEN + "[Info] Model Loaded from {}".format(model_path))
            return model
        elif model_path == None and scaler_path != None:
            scaler = load(open(scaler_path, "rb"))
            print(Fore.GREEN + "[Info] Scaler Loaded from {}".format(scaler_path))
            return scaler
        else:
            raise ValueError("No path specified.Please provide actual path\n")

    def imp_features(self, extensive=False, *args, **kwargs):
        """
        Returns the importance features of the dataset

        Args:

            extensive (bool): [If True shows the importance of all features exitensively and will take more time] [default = False]
            **args: [Additional arguments]
            **kwargs: [Additional keyword arguments]
        """
        if self.original_predictor == "all":
            raise TypeError(
                "[Error] This method is only applicable on single predictor"
            )
        if not extensive:
            self.preprocess.permutational_feature_imp(
                self.features, self.X_train, self.y_train, model=self.regressor
            )
        if extensive:
            self.preprocess.shap_feature_imp(
                self.features, self.X_train, model=self.regressor, *args, **kwargs
            )



import copy
import os
from re import S
import time
import warnings

import numpy as np
import optuna
import pandas as pd
from IPython.display import display
from joblib import dump, load
from luciferml.supervised.utils.best import Best
from luciferml.supervised.utils.configs import *
from luciferml.supervised.utils.predictors import classification_predictor
from luciferml.supervised.utils.preprocesser import PreProcesser
from luciferml.supervised.utils.tuner.luciferml_tuner import luciferml_tuner
from luciferml.supervised.utils.validator import *
from optuna.samplers._tpe.sampler import TPESampler
from sklearn.metrics import accuracy_score
from colorama import Fore


class Classification:
    def __init__(
        self,
        predictor=["lr"],
        params={},
        tune=False,
        test_size=0.2,
        cv_folds=10,
        random_state=42,
        pca_kernel="linear",
        n_components_lda=1,
        lda="n",
        pca="n",
        n_components_pca=2,
        metrics=[
            "accuracy",
        ],
        loss="binary_crossentropy",
        validation_split=0.20,
        tune_mode=1,
        smote="n",
        k_neighbors=1,
        verbose=False,
        exclude_models=[],
        path=None,
        optuna_sampler=TPESampler(multivariate=True),
        optuna_direction="maximize",
        optuna_n_trials=100,
        optuna_metric="accuracy",
        lgbm_objective="binary",
    ):
        """
        Encode Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2 , scales self.X_train, self.X_val using StandardScaler.\n
        Fits every model on training set and predicts results find and plots Confusion Matrix,\n
        finds accuracy of model applies K-Fold Cross Validation\n
        and stores accuracy in variable name accuracy and model name in self.classifier name and returns both as a tuple.\n
        Applies HyperParam Tuning and gives best params and accuracy.\n

        Parameters:

            features : array
                        features array
            lables : array
                        labels array
            predictor : list
                        Predicting model to be used
                        Default ['lr']  - Logistic Regression\n
                        Available Predictors:
                                lr - Logisitic Regression\n
                                sgd - Stochastic Gradient Descent Classifier\n
                                perc - Perceptron\n
                                pass - Passive Aggressive Classifier\n
                                ridg - Ridge Classifier\n
                                svm -SupportVector Machine\n
                                knn - K-Nearest Neighbours\n
                                nb - GaussianNaive bayes\n
                                rfc- Random Forest self.Classifier\n
                                gbc - Gradient Boosting Classifier\n
                                ada - AdaBoost Classifier\n
                                bag - Bagging Classifier\n
                                extc - Extra Trees Classifier\n
                                lgbm - LightGBM Classifier\n
                                cat - CatBoost Classifier\n
                                xgb- XGBoost self.Classifier\n
                                ann - MultiLayer Perceptron Classifier\n
                                all - Applies all above classifiers\n
                                
            params : dict
                        contains parameters for model
            tune : boolean
                    when True Applies GridSearch CrossValidation
                    Default is False
            test_size: float or int, default=.2
                        If float, should be between 0.0 and 1.0 and represent
                        the proportion of the dataset to include in
                        the test split.
                        If int, represents the absolute number of test samples.
            cv_folds : int
                    No. of cross validation folds. Default = 10
            pca : str
                if 'y' will apply PCA on Train and Validation set. Default = 'n'
            lda : str
                if 'y' will apply LDA on Train and Validation set. Default = 'n'
            pca_kernel : str
                    Kernel to be use in PCA. Default = 'linear'
            n_components_lda : int
                    No. of components for LDA. Default = 1
            n_components_pca : int
                    No. of components for PCA. Default = 2
            loss : str
                    loss method for ann. Default = 'binary_crossentropy'
                    rate for dropout layer. Default = 0
            smote : str,
                Whether to apply SMOTE. Default = 'y'
            k_neighbors : int
                No. of neighbours for SMOTE. Default = 1
            verbose : boolean
                Verbosity of models. Default = False
            exclude_models : list
                List of models to be excluded when using predictor = 'all' . Default = []
            path : list
                List containing path to saved model and scaler. Default = None
                Example: [model.pkl, scaler.pkl]
            random_state : int
                Random random_state for reproducibility. Default = 42
            optuna_sampler : Function
                Sampler to be used in optuna. Default = TPESampler()
            optuna_direction : str
                Direction of optimization. Default = 'maximize'
                Available Directions:
                    maximize : Maximize
                    minimize : Minimize
            optuna_n_trials : int
                No. of trials for optuna. Default = 100
            optuna_metric: str
                Metric to be used in optuna. Default = 'r2'
            lgbm_objective : str
                Objective for lgbm classifier. Default = 'binary'

        Returns:

            Dict Containing Name of Classifiers, Its K-Fold Cross Validated Accuracy and Prediction set

            Dataframe containing all the models and their accuracies when predictor is 'all'

        Example:

            from luciferml.supervised.classification import Classification

            dataset = pd.read_csv('Social_Network_Ads.csv')

            X = dataset.iloc[:, :-1]

            y = dataset.iloc[:, -1]

            classifier = Classification(predictor = 'lr')

            classifier.fit(X, y)

            result = classifier.result()

        """
        self.preprocess = PreProcesser()
        if type(predictor) == list:
            if not "all" in predictor:
                self.predictor = predictor[0] if len(
                    predictor) == 1 else predictor
            else:
                self.predictor = predictor
        else:
            self.predictor = predictor
        bool_pred, pred = pred_check(predictor, pred_type="classification")
        if not bool_pred:
            raise ValueError(unsupported_pred_warning.format(pred))
        self.original_predictor = predictor
        self.params = params
        self.tune = tune
        self.test_size = test_size
        self.cv_folds = cv_folds
        self.random_state = random_state
        self.pca_kernel = pca_kernel
        self.n_components_lda = n_components_lda
        self.lda = lda
        self.pca = pca
        self.n_components_pca = n_components_pca
        self.metrics = metrics
        self.loss = loss
        self.validation_split = validation_split
        self.tune_mode = tune_mode
        self.rerun = False
        self.smote = smote
        self.k_neighbors = k_neighbors
        self.verbose = verbose
        self.exclude_models = exclude_models
        self.sampler = optuna_sampler
        self.direction = optuna_direction
        self.n_trials = optuna_n_trials
        self.metric = optuna_metric
        self.lgbm_objective = lgbm_objective

        self.accuracy_scores = {}
        self.reg_result = {}
        self.accuracy = 0
        self.y_pred = []
        self.kfold_accuracy = 0
        self.classifier_name = ""
        self.sc = 0

        self.kfoldacc = []
        self.acc = []
        self.bestacc = []
        self.bestparams = []
        self.tuned_trained_model = []
        self.best_classifier_path = ""
        self.scaler_path = ""
        self.classifier_model = []
        self.result_df = pd.DataFrame(index=None)
        self.classifiers = copy.deepcopy(classifiers)
        for i in self.exclude_models:
            self.classifiers.pop(i)
        self.best_classifier = "First Run the Predictor in All mode"
        self.objective = None
        self.pred_mode = ""
        self.model_to_predict = []

        if path != None:
            try:
                self.classifier, self.sc = self.__load(path)
            except Exception as e:
                print(Fore.RED + e)
                print(Fore.RED + "Model not found")
        if not self.verbose:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

    def fit(self, features, labels):
        """[Takes Features and Labels and Encodes Categorical Data then Applies SMOTE , Splits the features and labels in training and validation sets with test_size = .2
        scales X_train, self.X_val using StandardScaler.
        Fits every model on training set and predicts results,
        finds accuracy of model applies K-Fold Cross Validation
        and stores its accuracies in a dictionary containing Model name as Key and accuracies as values and returns it
        Applies GridSearch Cross Validation and gives best params out from param list.]

        Args:
            features ([Pandas DataFrame]): [DataFrame containing Features]
            labels ([Pandas DataFrame]): [DataFrame containing Labels]
        """
        self.features = features
        self.labels = labels
        # Time Function ---------------------------------------------------------------------

        self.start = time.time()
        print(Fore.MAGENTA + intro, "\n")
        print(Fore.GREEN + "Started LuciferML [", "\u2713", "]\n")
        if not self.rerun:
            # CHECKUP ---------------------------------------------------------------------
            if not isinstance(self.features, pd.DataFrame) and not isinstance(
                self.labels, pd.Series
            ):
                print(
                    Fore.RED
                    + "TypeError: This Function take features as Pandas Dataframe and labels as Pandas Series. Please check your implementation.\n"
                )
                self.end = time.time()
                print(self.end - self.start)
                return

            print(Fore.YELLOW + "Preprocessing Started [*]\n")
            self.features, self.labels = self.preprocess.encoder(
                self.features, self.labels
            )

            self.features, self.labels = sparse_check(self.features, self.labels)

            (
                self.X_train,
                self.X_val,
                self.y_train,
                self.y_val,
                self.sc,
            ) = self.preprocess.data_preprocess(
                self.features,
                self.labels,
                self.test_size,
                self.random_state,
                self.smote,
                self.k_neighbors,
            )

            self.X_train, self.X_val = self.preprocess.dimensionality_reduction(
                self.lda,
                self.pca,
                self.X_train,
                self.X_val,
                self.y_train,
                self.n_components_lda,
                self.n_components_pca,
                self.pca_kernel,
                self.start,
            )

        print(Fore.GREEN + "Preprocessing Done [", "\u2713", "]\n")

        if self.original_predictor == "all" or type(self.predictor) == list:
            if 'all' in self.predictor and type(self.predictor)==list:
                self.predictor.remove('all')
            self.model_to_predict = (
                self.predictor if len(self.predictor) > 1 and type(self.predictor) == list else self.classifiers
            )
            
            self.result_df["Name"] = (
                list(self.classifiers[i] for i in self.predictor)
                if type(self.predictor) == list and len(self.predictor) > 1
                else list(self.classifiers.values())
            )
            self.pred_mode = "all" if len(self.predictor) > 1 and type(
                self.predictor) == list else "single"
            self.__fitall()
            return

        self.classifier, self.objective = classification_predictor(
            self.predictor,
            self.params,
            self.X_train,
            self.y_train,
            self.cv_folds,
            self.random_state,
            self.metric,
            verbose=self.verbose,
            lgbm_objective=self.lgbm_objective,
        )
        try:
            self.classifier.fit(self.X_train, self.y_train)
        except Exception as error:
            print(Fore.RED + "Classifier Build Failed with error: ", error, "\n")
        finally:
            print(Fore.GREEN + "Model Trained Successfully [", "\u2713", "]\n")

        try:
            print(Fore.YELLOW + "Evaluating Model Performance [*]\n")
            self.y_pred = self.classifier.predict(self.X_val)
            if self.predictor == "ann":
                self.y_pred = (self.y_pred > 0.5).astype("int32")
            self.accuracy = accuracy_score(self.y_val, self.y_pred)
            print(Fore.CYAN + "        Validation Accuracy is : {:.2f} %".format(self.accuracy * 100))
            self.classifier_name, self.kfold_accuracy = kfold(
                self.classifier,
                self.predictor,
                self.X_train,
                self.y_train,
                self.cv_folds,
            )
            self.preprocess.confusion_matrix(self.y_pred, self.y_val)
        except Exception as error:
            print(Fore.RED + "Model Evaluation Failed with error: ", error, "\n")
        finally:
            print(Fore.GREEN + "Model Evaluation Completed [", "\u2713", "]\n")

        if not self.predictor == "nb" and self.tune:
            self.__tuner()

        print(Fore.GREEN + "Completed LuciferML Run [", "\u2713", "]\n")
        self.end = time.time()
        final_time = self.end - self.start
        print(Fore.BLUE + "Time Elapsed : ", f"{final_time:.2f}", "seconds \n")

    def __fitall(self):
        print(Fore.YELLOW + "Training LuciferML [*]\n")
        if self.params != {}:
            warnings.warn(params_use_warning, UserWarning)
            self.params = {}
        for _, self.predictor in enumerate(self.model_to_predict):
            if not self.predictor in self.exclude_models:
                try:
                    self.classifier, self.objective = classification_predictor(
                    self.predictor,
                    self.params,
                    self.X_train,
                    self.y_train,
                    self.cv_folds,
                    self.random_state,
                    self.metric,
                    mode="multi",
                    verbose=self.verbose,
                    lgbm_objective=self.lgbm_objective,
                )
                except Exception as error:
                    print(
                        Fore.RED + classifiers[self.predictor],
                        "Model Train Failed with error: ",
                        error,
                        "\n",
                    )
                try:
                    self.classifier.fit(self.X_train, self.y_train)
                    self.y_pred = self.classifier.predict(self.X_val)
                    if self.predictor == "ann":
                        self.y_pred = (self.y_pred > 0.5).astype("int32")
                    self.accuracy = accuracy_score(self.y_val, self.y_pred)
                    self.acc.append(self.accuracy * 100)
                    self.classifier_name, self.kfold_accuracy = kfold(
                        self.classifier,
                        self.predictor,
                        self.X_train,
                        self.y_train,
                        self.cv_folds,
                        all_mode=True,
                    )
                    self.kfoldacc.append(self.kfold_accuracy)
                    self.classifier_model.append(self.classifier)
                except Exception as error:
                    print(
                        classifiers[self.predictor],
                        "Evaluation Failed with error: ",
                        error,
                        "\n",
                    )
                if self.tune:
                    self.__tuner(all_mode=True, single_mode=False)

        self.result_df["Accuracy"] = self.acc
        self.result_df["KFold Accuracy"] = self.kfoldacc
        self.result_df["Model"] = self.classifier_model
        if self.tune:
            self.result_df["Best Parameters"] = self.bestparams
            self.result_df["Best Accuracy"] = self.bestacc
            self.best_classifier = Best(
                self.result_df.loc[self.result_df["Best Accuracy"].idxmax()],
                self.tune,
            )
        else:
            self.best_classifier = Best(
                self.result_df.loc[self.result_df["KFold Accuracy"].idxmax()], self.tune
            )
        print(Fore.GREEN + "Training Done [", "\u2713", "]\n")
        print(Fore.CYAN + "Results Below\n")
        display(self.result_df)
        print(Fore.GREEN + "\nCompleted LuciferML Run [", "\u2713", "]\n")
        if len(self.model_to_predict) > 1:
            self.best_classifier_path, self.scaler_path = self.save(
                best=True, model=self.best_classifier.model, scaler=self.sc
            )
            print(
                Fore.CYAN
                + "Saved Best Model to {} and its scaler to {}".format(
                    self.best_classifier_path, self.scaler_path
                ),
                "\n",
            )
        self.end = time.time()
        final_time = self.end - self.start
        print(Fore.BLUE + "Time Elapsed : ", f"{final_time:.2f}", "seconds \n")
        return

    def __tuner(self, all_mode=False, single_mode=False):
        if not all_mode:
            print(Fore.YELLOW + "Tuning Started [*]\n")
        if not self.predictor == "nb":
            (
                self.best_params,
                self.best_accuracy,
                self.best_trained_model,
            ) = luciferml_tuner(
                self.predictor,
                self.objective,
                self.n_trials,
                self.sampler,
                self.direction,
                self.X_train,
                self.y_train,
                self.cv_folds,
                self.random_state,
                self.metric,
                all_mode=all_mode,
            )
        if self.predictor == "nb":
            self.best_params = "Not Applicable"
            self.best_accuracy = 0
        self.bestparams.append(self.best_params)
        self.bestacc.append(self.best_accuracy * 100)
        self.tuned_trained_model.append(self.best_trained_model)
        if not all_mode or single_mode:
            print(Fore.GREEN + "Tuning Done [", "\u2713", "]\n")

    def result(self):
        """[Makes a dictionary containing Classifier Name, K-Fold CV Accuracy, RMSE, Prediction set.]

        Returns:
            [dict]: [Dictionary containing :
                        - "Classifier" - Classifier Name
                        - "Accuracy" - KFold CV Accuracy
                        - "YPred" - Array for Prediction set
                        ]
            [dataframe] : [Dataset containing accuracy and best_params
                            for all predictors only when predictor = 'all' is used
                            ]
        """
        if not self.pred_mode == "all":
            self.reg_result["Classifier"] = self.classifier_name
            self.reg_result["Accuracy"] = self.kfold_accuracy
            self.reg_result["YPred"] = self.y_pred

            return self.reg_result
        if self.pred_mode == "all":
            return self.result_df

    def predict(self, X_test):
        """[Takes test set and returns predictions for that test set]

        Args:
            X_test ([Array]): [Array Containing Test Set]

        Returns:
            [Array]: [Predicted set for given test set]
        """
        if not self.pred_mode == "all":
            X_test = np.array(X_test)
            if X_test.ndim == 1:
                X_test = X_test.reshape(1, -1)

            y_test = self.classifier.predict(self.sc.transform(X_test))

            return y_test
        if self.pred_mode == "all":
            raise TypeError("Predict is only applicable on single predictor")

    def save(self, path=None, best=False, **kwargs):
        """
        Saves the model and its scaler to a file provided with a path.
        If no path is provided will create a directory named
        lucifer_ml_info/models/ and lucifer_ml_info/scaler/ in current working directory
        Args:
            path ([list]): [List containing path to save the model and scaler.]
                Example: path = ["model.pkl", "scaler.pkl"]

        Returns:
            Path to the saved model and its scaler.
        """
        if not type(path) == list and path != None:
            raise TypeError("Path must be a list")
        if self.pred_mode == "all" and best == False:
            raise TypeError("Cannot save model for all predictors")
        dir_path_model = path[0] if path else "lucifer_ml_info/models/classifier/"
        dir_path_scaler = path[1] if path else "lucifer_ml_info/scalers/classifier/"
        model_name = classifiers[self.predictor].replace(" ", "_")
        if best:
            dir_path_model = "lucifer_ml_info/best/classifier/models/"
            dir_path_scaler = "lucifer_ml_info/best/classifier/scalers/"
            model_name = self.best_classifier.name.replace(" ", "_")
        os.makedirs(dir_path_model, exist_ok=True)
        os.makedirs(dir_path_scaler, exist_ok=True)
        timestamp = str(int(time.time()))
        path_model = dir_path_model + model_name + "_" + timestamp + ".pkl"
        path_scaler = (
            dir_path_scaler + model_name + "_" + "Scaler" + "_" + timestamp + ".pkl"
        )
        if (
            not kwargs.get("model")
            and not kwargs.get("best")
            and not kwargs.get("scaler")
        ):
            dump(self.classifier, open(path_model, "wb"))
            dump(self.sc, open(path_scaler, "wb"))
        else:
            dump(kwargs.get("model"), open(path_model, "wb"))
            dump(kwargs.get("scaler"), open(path_scaler, "wb"))
        if not best:
            print("Model Saved at {} and Scaler at {}".format(path_model, path_scaler))
        return path_model, path_scaler

    def __load(self, path=None):
        """
        Loads model and scaler from the specified path
        Args:
            path ([list]): [List containing path to load the model and scaler.]
                Example: path = ["model.pkl", "scaler.pkl"]

        Returns:
            [Model] : [Loaded model]
            [Scaler] : [Loaded scaler]
        """

        model_path = path[0] if path[0] else None
        scaler_path = path[1] if path[1] else None
        if not ".pkl" in model_path and not model_path == None:
            raise TypeError(
                "[Error] Model Filetype not supported. Please use .pkl type "
            )
        if not ".pkl" in scaler_path and not scaler_path == None:
            raise TypeError(
                "[Error] Scaler Filetype not supported. Please use .pkl type "
            )
        if model_path != None and scaler_path != None:
            model = load(open(model_path, "rb"))
            scaler = load(open(scaler_path, "rb"))
            print(
                Fore.GREEN
                + "[Info] Model and Scaler Loaded from {} and {}".format(
                    model_path, scaler_path
                )
            )
            return model, scaler
        elif model_path != None and scaler_path == None:
            model = load(open(model_path, "rb"))
            print(Fore.GREEN + "[Info] Model Loaded from {}".format(model_path))
            return model
        elif model_path == None and scaler_path != None:
            scaler = load(open(scaler_path, "rb"))
            print(Fore.GREEN + "[Info] Scaler Loaded from {}".format(scaler_path))
            return scaler
        else:
            raise ValueError("No path specified.Please provide actual path\n")

    def imp_features(self, extensive=False, *args, **kwargs):
        """
        Returns the importance features of the dataset

        Args:

            extensive (bool): [If True shows the importance of all features exitensively and will take more time] [default = False]
            **args: [Additional arguments]
            **kwargs: [Additional keyword arguments]
        """
        if self.original_predictor == "all":
            raise TypeError(
                "[Error] This method is only applicable on single predictor"
            )
        if not extensive:
            self.preprocesspermutational_feature_imp(
                self.features, self.X_train, self.y_train, model=self.classifier
            )
        if extensive:
            self.preprocessshap_feature_imp(
                self.features, self.X_train, model=self.classifier, *args, **kwargs
            )



import optuna


class Tuner:
    def __init__(self, sampler, n_trials=100, direction="maximize"):
        self.n_trials = n_trials
        self.sampler = sampler
        self.direction = direction

    def tune(self, objective):
        study = optuna.create_study(direction=self.direction, sampler=self.sampler)
        study.optimize(objective, n_trials=self.n_trials, n_jobs=-1, gc_after_trial=True)
        params = study.best_params
        best_score = study.best_value
        return params, best_score



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="lucifer-ml",
    packages=[
        "luciferml",
        "luciferml.supervised",
        "luciferml.supervised.utils",
        "luciferml.supervised.utils.tuner",
        "luciferml.supervised.utils.tuner.optuna",
        "luciferml.supervised.utils.tuner.optuna.objectives",
    ],
    version="0.0.81a",
    license="MIT",
    description="Automated ML by d4rk-lucif3r",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Arsh Anwar",
    author_email="lucifer78908@gmail.com",
    url="https://github.com/d4rk-lucif3r/LuciferML",
    keywords=["luciferML", "AutoML", "Python"],
    install_requires=open("requirements.txt").readlines(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



from sklearn.model_selection import cross_val_score
import scipy
from luciferml.supervised.utils.configs import *
from colorama import Fore


def pred_check(predictor, pred_type):
    if pred_type == "regression":
        avlbl_predictors = list(regressors_ver.keys())
    elif pred_type == "classification":
        avlbl_predictors = list(classifiers_ver.keys())
    if type(predictor) == str:
        if predictor in avlbl_predictors:
            return True, predictor
        else:
            return False, predictor
    elif type(predictor) == list:
        for i in predictor:
            if i not in avlbl_predictors:
                return False, i
        return True, None


def sparse_check(features, labels):
    features = features
    labels = labels
    """
        Takes features and labels as input and checks if any of those is sparse csr_matrix.
        """
    try:
        if scipy.sparse.issparse(features[()]):
            features = features[()].toarray()
        elif scipy.sparse.issparse(labels[()]):
            labels = labels[()].toarray()
    except Exception as error:
        pass
    return (features, labels)


def kfold(model, predictor, X_train, y_train, cv_folds, isReg=False, all_mode=False):
    """
    Takes predictor, input_units, epochs, batch_size, X_train, y_train, cv_folds, and accuracy_scores dictionary.
    Performs K-Fold Cross validation and stores result in accuracy_scores dictionary and returns it.
    """
    if not isReg:
        name = classifiers
        scoring = "accuracy"
    if isReg:
        name = regressors
        scoring = "r2"
    try:
        accuracies = cross_val_score(
            estimator=model, X=X_train, y=y_train, cv=cv_folds, scoring=scoring
        )
        if not all_mode:
            if not isReg:
                print("        KFold Accuracy: {:.2f} %".format(accuracies.mean() * 100))
            if isReg:
                print("        R2 Score: {:.2f} %".format(accuracies.mean() * 100))
        model_name = name[predictor]
        accuracy = accuracies.mean() * 100
        if not all_mode:
            print(
                "        Standard Deviation: {:.2f} %".format(accuracies.std() * 100),
                "\n",
            )
        return (model_name, accuracy)

    except Exception as error:
        print(Fore.RED + "K-Fold Cross Validation failed with error: ", error, "\n")



from colorama import Fore


class Best:
    """
    Best is used to utilise the best model when predictor = 'all' is used.

    """

    def __init__(self, best_model, tune, isReg=False):
        self.__best_model = best_model
        self.model = self.__best_model["Model"]
        self.name = self.__best_model["Name"]
        self.tune = tune
        if isReg:
            self.r2_score = self.__best_model["R2 Score"]
            self.mae = self.__best_model["Mean Absolute Error"]
            self.rmse = self.__best_model["Root Mean Squared Error"]
        if not isReg:
            self.accuracy = self.__best_model["Accuracy"]
        self.kfold_acc = self.__best_model["KFold Accuracy"]
        if tune == True:
            self.best_params = self.__best_model["Best Parameters"]
            self.best_accuracy = self.__best_model["Best Accuracy"]
        else:
            self.best_params = "Run with tune = True to get best parameters"
        self.isReg = isReg

    def summary(self):
        """Returns a summary of the best model"""
        print("\nBest Model Summary:")
        print(Fore.CYAN + "Name: ", self.name)
        if self.isReg:
            print(Fore.CYAN + "R2 Score: ", self.r2_score)
            print(Fore.CYAN + "Mean Absolute Error: ", self.mae)
            print(Fore.CYAN + "Root Mean Squared Error: ", self.rmse)
        else:
            print(Fore.CYAN + "Accuracy: ", self.accuracy)
        print(Fore.CYAN + "KFold Accuracy: ", self.kfold_acc)
        if self.tune:
            print(Fore.CYAN + "Best Parameters: ", self.best_params)
            print(Fore.CYAN + "Best Accuracy: ", self.best_accuracy)
        print("\n")

    def predict(self, pred):
        """Predicts the output of the best model"""
        prediction = self.__best_model["Model"].predict(pred)
        return prediction



from catboost import CatBoostRegressor
from lightgbm import LGBMRegressor
from sklearn.ensemble import (
    AdaBoostRegressor,
    BaggingRegressor,
    ExtraTreesRegressor,
    GradientBoostingRegressor,
    RandomForestRegressor,
)
from sklearn.kernel_ridge import KernelRidge
from sklearn.linear_model import (
    BayesianRidge,
    ElasticNet,
    LinearRegression,
    SGDRegressor,
)
from sklearn.model_selection import cross_val_score
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.svm import SVR
from sklearn.tree import DecisionTreeRegressor
from xgboost import XGBRegressor


class RegressionObjectives:
    def __init__(self, X, y, cv=5, random_state=42, metric="r2"):
        self.metric = metric
        self.cv = cv
        self.X = X
        self.y = y
        self.random_state = random_state

    def lin_regressor_objective(self, trial):
        param = {
            "fit_intercept": trial.suggest_categorical("fit_intercept", [True, False]),
            "copy_X": trial.suggest_categorical("copy_X", [True, False]),
        }
        regressor = LinearRegression(**param, n_jobs=-1)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, scoring=self.metric
        )
        return scores.mean()

    def sgd_regressor_objective(self, trial):
        param = {
            "loss": trial.suggest_categorical(
                "loss", ["squared_loss", "huber", "epsilon_insensitive"]
            ),
            "penalty": trial.suggest_categorical(
                "penalty", ["none", "l2", "l1", "elasticnet"]
            ),
            "alpha": trial.suggest_float("alpha", 1e-10, 1e-3),
            "l1_ratio": trial.suggest_float("l1_ratio", 0.0, 1.0),
            "learning_rate": trial.suggest_categorical(
                "learning_rate", ["constant", "optimal", "invscaling", "adaptive"]
            ),
            "eta0": trial.suggest_float("eta0", 0.0, 1.0),
            "power_t": trial.suggest_float("power_t", 0.0, 1.0),
            "warm_start": trial.suggest_categorical("warm_start", [True, False]),
            "average": trial.suggest_categorical("average", [True, False]),
            "random_state": self.random_state,
        }
        regressor = SGDRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def krr_regressor_objective(self, trial):
        param = {
            "alpha": trial.suggest_loguniform("alpha", 1e-10, 1e-3),
            "kernel": trial.suggest_categorical("kernel", ["linear", "rbf"]),
            "degree": trial.suggest_int("degree", 1, 3),
            "gamma": trial.suggest_loguniform("gamma", 1e-10, 1e-3),
            "coef0": trial.suggest_loguniform("coef0", 1e-10, 1e-3),
        }
        regressor = KernelRidge(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def elas_regressor_objective(self, trial):
        param = {
            "alpha": trial.suggest_loguniform("alpha", 1e-10, 1e-3),
            "l1_ratio": trial.suggest_float("l1_ratio", 0.0, 1.0),
            "max_iter": trial.suggest_int("max_iter", 100, 1000),
            "selection": trial.suggest_categorical("selection", ["cyclic", "random"]),
            "tol": trial.suggest_loguniform("tol", 1e-10, 1e-3),
            "random_state": self.random_state,
        }
        regressor = ElasticNet(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def br_regressor_objective(self, trial):
        param = {
            "alpha_1": trial.suggest_loguniform("alpha_1", 1e-10, 1e-3),
            "alpha_2": trial.suggest_loguniform("alpha_2", 1e-10, 1e-3),
            "lambda_1": trial.suggest_loguniform("lambda_1", 1e-10, 1e-3),
            "lambda_2": trial.suggest_loguniform("lambda_2", 1e-10, 1e-3),
            "fit_intercept": trial.suggest_categorical("fit_intercept", [True, False]),
            "normalize": trial.suggest_categorical("normalize", [True, False]),
            "copy_X": trial.suggest_categorical("copy_X", [True, False]),
        }
        regressor = BayesianRidge(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def svr_regressor_objective(self, trial):
        param = {
            "C": trial.suggest_loguniform("C", 1e-10, 1e-3),
            "kernel": trial.suggest_categorical("kernel", ["linear", "rbf"]),
            "degree": trial.suggest_int("degree", 1, 3),
            "gamma": trial.suggest_loguniform("gamma", 1e-10, 1e-3),
            "coef0": trial.suggest_loguniform("coef0", 1e-10, 1e-3),
            "shrinking": trial.suggest_categorical("shrinking", [True, False]),
            "tol": trial.suggest_loguniform("tol", 1e-10, 1e-3),
            "cache_size": trial.suggest_loguniform("cache_size", 1e-10, 1e-3),
            "verbose": trial.suggest_categorical("verbose", [True, False]),
            "max_iter": trial.suggest_int("max_iter", 100, 1000),
        }
        regressor = SVR(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def knr_regressor_objective(self, trial):
        param = {
            "n_neighbors": trial.suggest_int("n_neighbors", 1, 10),
            "weights": trial.suggest_categorical("weights", ["uniform", "distance"]),
            "algorithm": trial.suggest_categorical(
                "algorithm", ["auto", "ball_tree", "kd_tree", "brute"]
            ),
            "leaf_size": trial.suggest_int("leaf_size", 1, 100),
            "p": trial.suggest_int("p", 1, 3),
            "n_jobs": -1,
        }
        regressor = KNeighborsRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def dt_regressor_objective(self, trial):
        param = {
            "criterion": trial.suggest_categorical(
                "criterion", ["mse", "friedman_mse", "mae"]
            ),
            "splitter": trial.suggest_categorical("splitter", ["best", "random"]),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "min_samples_split": trial.suggest_int("min_samples_split", 2, 10),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 10),
            "min_weight_fraction_leaf": trial.suggest_loguniform(
                "min_weight_fraction_leaf", 1e-10, 1e-3
            ),
            "max_features": trial.suggest_categorical(
                "max_features", ["auto", "sqrt", "log2", None]
            ),
            "max_leaf_nodes": trial.suggest_int("max_leaf_nodes", 2, 10),
            "min_impurity_decrease": trial.suggest_loguniform(
                "min_impurity_decrease", 1e-10, 1e-3
            ),
            "random_state": self.random_state,
        }
        regressor = DecisionTreeRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def gbr_regressor_objective(self, trial):
        param = {
            "loss": trial.suggest_categorical(
                "loss", ["ls", "lad", "huber", "quantile"]
            ),
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-10, 1e-3),
            "n_estimators": trial.suggest_int("n_estimators", 10, 1000),
            "criterion": trial.suggest_categorical(
                "criterion", ["friedman_mse", "mae"]
            ),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "min_samples_split": trial.suggest_int("min_samples_split", 2, 10),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 10),
            "min_weight_fraction_leaf": trial.suggest_loguniform(
                "min_weight_fraction_leaf", 1e-10, 1e-3
            ),
            "max_features": trial.suggest_categorical(
                "max_features", ["auto", "sqrt", "log2", None]
            ),
            "max_leaf_nodes": trial.suggest_int("max_leaf_nodes", 2, 10),
            "min_impurity_decrease": trial.suggest_loguniform(
                "min_impurity_decrease", 1e-10, 1e-3
            ),
            "random_state": self.random_state,
        }
        regressor = GradientBoostingRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def ada_regressor_objective(self, trial):
        param = {
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-10, 1e-3),
            "n_estimators": trial.suggest_int("n_estimators", 10, 1000),
            "loss": trial.suggest_categorical(
                "loss", ["linear", "square", "exponential"]
            ),
            "random_state": self.random_state,
        }
        regressor = AdaBoostRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def bag_regressor_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 10, 1000),
            "bootstrap_features": trial.suggest_categorical(
                "bootstrap_features", [True, False]
            ),
            "oob_score": trial.suggest_categorical("oob_score", [True, False]),
            "max_samples": trial.suggest_uniform("max_samples", 0.0, 1.0),
            "max_features": trial.suggest_uniform("max_features", 0.0, 1.0),
            "random_state": self.random_state,
            "n_jobs": -1,
        }
        regressor = BaggingRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def extr_regressor_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 10, 1000),
            "criterion": trial.suggest_categorical(
                "criterion", ["mse", "friedman_mse", "mae"]
            ),
            "max_depth": trial.suggest_int("max_depth", 1, 10),
            "min_samples_split": trial.suggest_int("min_samples_split", 2, 10),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 10),
            "min_weight_fraction_leaf": trial.suggest_loguniform(
                "min_weight_fraction_leaf", 1e-10, 1e-3
            ),
            "max_features": trial.suggest_categorical(
                "max_features", ["auto", "sqrt", "log2", None]
            ),
            "max_leaf_nodes": trial.suggest_int("max_leaf_nodes", 2, 10),
            "min_impurity_decrease": trial.suggest_loguniform(
                "min_impurity_decrease", 1e-10, 1e-3
            ),
            "bootstrap": True,
            "oob_score": trial.suggest_categorical("oob_score", [True, False]),
            "random_state": self.random_state,
            "n_jobs": -1,
        }
        regressor = ExtraTreesRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def rfr_regressor_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 200, 1500),
            "max_features": trial.suggest_categorical("max_features", ["auto", "sqrt"]),
            "max_depth": trial.suggest_int("max_depth", 10, 80, log=True),
            "min_samples_split": trial.suggest_int("min_samples_split", 2, 15),
            "min_samples_leaf": trial.suggest_int("min_samples_leaf", 1, 9),
            "bootstrap": trial.suggest_categorical("bootstrap", [True, False]),
        }
        regressor = RandomForestRegressor(**param, n_jobs=-1, verbose=0)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def xgb_regressor_objective(self, trial):
        param = {
            "n_estimators": trial.suggest_int("n_estimators", 500, 4000),
            "max_depth": trial.suggest_int("max_depth", 8, 16),
            "min_child_weight": trial.suggest_int("min_child_weight", 1, 300),
            "gamma": trial.suggest_int("gamma", 1, 3),
            "learning_rate": 0.01,
            "colsample_bytree": trial.suggest_discrete_uniform(
                "colsample_bytree", 0.5, 1, 0.1
            ),
            "lambda": trial.suggest_loguniform("lambda", 1e-3, 10.0),
            "alpha": trial.suggest_loguniform("alpha", 1e-3, 10.0),
            "subsample": trial.suggest_categorical("subsample", [0.6, 0.7, 0.8, 1.0]),
            "random_state": 42,
        }
        regressor = XGBRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def cat_regressor_objective(self, trial):
        params = {
            "iterations": trial.suggest_int("iterations", 50, 300),
            "depth": trial.suggest_int("depth", 4, 10),
            "random_strength": trial.suggest_int("random_strength", 0, 100),
            "bagging_temperature": trial.suggest_loguniform(
                "bagging_temperature", 0.01, 100.00
            ),
            "learning_rate": trial.suggest_loguniform("learning_rate", 1e-3, 1e-1),
            "od_type": trial.suggest_categorical("od_type", ["IncToDec", "Iter"]),
        }
        regressor = CatBoostRegressor(**params)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def lgbm_regressor_objective(self, trial):

        param = {
            "boosting_type": "gbdt",
            "objective": "regression",
            "metric": "rmse",
            "learning_rate": trial.suggest_categorical(
                "learning_rate", [0.0125, 0.025, 0.05, 0.1]
            ),
            "num_leaves": trial.suggest_int("num_leaves", 2, 2048),
            "lambda_l1": trial.suggest_float("lambda_l1", 1e-8, 10.0, log=True),
            "lambda_l2": trial.suggest_float("lambda_l2", 1e-8, 10.0, log=True),
            "colsample_bytree": min(
                trial.suggest_float("colsample_bytree", 0.3, 1.0 + 1e-8), 1.0
            ),
            "bagging_fraction": min(
                trial.suggest_float("bagging_fraction", 0.3, 1.0 + 1e-8), 1.0
            ),
            "bagging_freq": trial.suggest_int("bagging_freq", 1, 7),
            "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 1, 100),
            "feature_pre_filter": False,
            "random_state": self.random_state,
            "num_threads": -1,
            "extra_trees": trial.suggest_categorical("extra_trees", [True, False]),
        }
        regressor = LGBMRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()

    def mlp_regressor_objective(self, trial):
        param = {
            "hidden_layer_sizes": trial.suggest_int("hidden_layer_sizes", 2, 10),
            "activation": trial.suggest_categorical("activation", ["logistic", "tanh"]),
            "solver": trial.suggest_categorical("solver", ["lbfgs", "adam"]),
            "alpha": trial.suggest_loguniform("alpha", 1e-8, 1e-1),
            "learning_rate": trial.suggest_categorical(
                "learning_rate", ["constant", "adaptive"]
            ),
            "max_iter": trial.suggest_int("max_iter", 1, 2000),
            "random_state": self.random_state,
            "verbose": 0,
            "early_stopping": True,
            "validation_fraction": 0.2,
            "n_iter_no_change": 10,
        }
        regressor = MLPRegressor(**param)
        scores = cross_val_score(
            regressor, self.X, self.y, cv=self.cv, n_jobs=-1, scoring=self.metric
        )
        return scores.mean()
